/*
 * Copyright (C) 1997 Red Hat Software, Inc.
 *	Cristian Gafton <gafton@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 	$Id: wmconfig.c,v 1.25 2000/07/10 19:43:34 gafton Exp $	 */

/*
 * WM config program
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <pwd.h>

/* static vars... */
static const char *ProgName = NULL;

/* What types of output we can do ... */
#define		DO_FVWM95	1	/* Fvwm95 output	*/
#define		DO_FVWM2	2	/* Fvwm2 output		*/
#define		DO_AFTERSTEP	3	/* Afterstep output	*/
#define		DO_MWM		4	/* Mwm output		*/
#define		DO_ICEWM	5	/* Icewm output		*/
#define		DO_WMAKER	6	/* wmaker output	*/
#define		DO_KDE		7	/* KDE output	*/
#define        	DO_DEBUG	9	/* Debugging		*/
#define		DO_BLACKBOX    10	/* Blackbox output	*/

static int output_type 	= 0;
unsigned int flags	= 0;

/* Includes our own stuff */
#include "wmconfig.h"

/*
 * This holds the root menu of our applications
 */
struct group *root_group = NULL;

/*
 * We read all the packages here, do overrides, then build
 * root_group from what's left
 */
extern struct package	*packages;
extern int		num_packages;

static const char *sys_confdir = NULL;
static const char *user_confdir = NULL;

/* variables to be shared with other modules */
const char 	*root_menu_name = NULL; /* name of the root menu */
const char  *output_directory = NULL; /* dirctory where put menu */

static node	*groups_ok, *groups_ignore;
		/*list of group names that are allowed to be output*/

/*
 * Prints out some help for the clueless looser...
 */
static void usage(void)
{
    printf("Usage: \t%s [options] [file ...]\nwhere options can be:\n", ProgName);
    printf("  --output\tdebug | fvwm95 | fvwm2 | afterstep | mwm |\n"
	   "\t\tanotherlevel | icewm | blackbox | wmaker | kde\n");
    printf("\t\t\t(gives the type of output; default is debug)\n");
    printf("  --sysdir\tsys_config_dir\t\t(default: \"%s\")\n",
	   DEFAULT_SYS_CONFDIR?DEFAULT_SYS_CONFDIR:"");
    printf("  --userdir\tuser_config_dir\t\t(default: \"$HOME/%s\")\n",
	   DEFAULT_USER_CONFDIR?DEFAULT_USER_CONFDIR:"");
    printf("  --rootmenu\troot_menu_name\t\t(default: \"%s\")\n",
	   DEFAULT_ROOT_NAME?DEFAULT_ROOT_NAME:"");
    printf("  --outputdir\toutput_directory\t(default: \"%s\")\n",
	   DEFAULT_OUTPUTDIR?"$HOME/"DEFAULT_OUTPUTDIR:"");
    printf("  --allow\tgroupspec\t\t\t(no defaults)\n");
    printf("  --ignore\tgroupspec\t\t\t(no defaults)\n");
    printf("  --flag\tno-icons,no-mini-icons,promote,\n");
    printf("\t\tdirectories,newstyle-directories\t(no defaults)\n");
    printf("  --check-existence\n");
    printf("  --no-icons\n");
    printf("  --no-mini-icons\n");
    printf("  --directories\n");
    printf("  --newstyle-directories\n");
    printf("  --no-promote\n");
    printf("  --no-sysdir\n");
    printf("  --no-userdir\n");
    printf("  --help\n");
    printf("  --version\n");
    exit(-1);
}

static void version(void)
{
#ifdef VERSION
    printf("%s version %s\n", ProgName, VERSION);
#else
    printf("%s (developer release)\n", ProgName);
#endif
    printf("$Id: wmconfig.c,v 1.25 2000/07/10 19:43:34 gafton Exp $\n");
    exit(0);
}

static int parse_options(int argc, char **argv)
{
    int option_index = 0;
    int c;
    static int val = 0;
    
    static struct option options[] = {
	{ "output",	required_argument, 	0, 'o' },
	{ "sysdir",	required_argument, 	0, 's' },
	{ "userdir",	required_argument, 	0, 'u' },
	{ "rootmenu",	required_argument, 	0, 'r' },
	{ "flags",	required_argument, 	0, 'f' },
	{ "help",	no_argument,		0, 'h' },
	{ "version",	no_argument,		0, 'v' },
	{ "no-promote",	no_argument,		&val, 	FLAG_NO_PROMOTE },
	{ "directories", no_argument,		&val, 	FLAG_DIRECTORIES },
	{ "newstyle-directories", no_argument,		&val, 	FLAG_NEWSTYLE_DIRECTORIES },
	{ "no-icons",	 no_argument,		&val, 	FLAG_NO_ICONS },
	{ "check-existence", no_argument,	&val, 	FLAG_CHECK_EXISTENCE },
	{ "no-mini-icons", no_argument,		&val, 	FLAG_NO_MINI_ICONS },
	{ "no-sysdir", 	no_argument,		&val, 	FLAG_NO_SYSDIR },
	{ "no-userdir", no_argument,		&val, 	FLAG_NO_USERDIR },
	{ "outputdir", required_argument,      0,    'd'},
	{ "allow",	required_argument,	0, 'a' },
	{ "ignore",	required_argument,	0, 'i' },	
	{0, 0, 0, 0 },
    };

    while ((c = getopt_long(argc, argv, "o:s:u:hr:f:d:a:i:", options,
			    &option_index)) != EOF) {
	switch (c) {
	    case 0:
		flags |= val;
		val = 0;
		break;
		
	    case 'o':
		if (optarg == NULL) {
		    usage();
		    break;
		}
		if (strncasecmp(optarg, "fvwm95", 6) == 0) 
		    output_type = DO_FVWM95;
		if (strncasecmp(optarg, "afterstep", 9) == 0)
		    output_type = DO_AFTERSTEP;
		if (strncasecmp(optarg, "mwm", 3) == 0) 
		    output_type = DO_MWM;
		if (strncasecmp(optarg, "wmaker", 6) == 0) 
		    output_type = DO_WMAKER;
		if (strncasecmp(optarg, "fvwm2", 5) == 0) 
		    output_type = DO_FVWM2;
		if (strncasecmp(optarg, "icewm", 5) == 0)
		    output_type = DO_ICEWM;
		if (strncasecmp(optarg, "blackbox", 8) == 0)
		    output_type = DO_BLACKBOX;
		if (strncasecmp(optarg, "kde", 3) == 0)
		    output_type = DO_KDE;
		if (strncasecmp(optarg, "debug", 5) == 0) 
		    output_type = DO_DEBUG;
		break;
	    case 's':
		if (optarg == NULL)
		    usage();
		sys_confdir = optarg;
		break;
	    case 'u':
		if (optarg == NULL)
		    usage();
		user_confdir = optarg;
		break;		
	    case 'r':
		if (optarg == NULL)
		    usage();
		root_menu_name = optarg;
		break;
	    case 'f':
		if (strstr(optarg, "no-icons") != NULL)
		    flags |= FLAG_NO_ICONS;
		if (strstr(optarg, "check-existence") != NULL)
		    flags |= FLAG_CHECK_EXISTENCE;
		if (strstr(optarg, "no-mini-icons") != NULL)
		    flags |= FLAG_NO_MINI_ICONS;
		if (strstr(optarg, "directories") != NULL)
		    flags |= FLAG_DIRECTORIES;
		if (strstr(optarg, "newstyle-directories") != NULL)
		    flags |= FLAG_NEWSTYLE_DIRECTORIES;
		if (strstr(optarg, "no-promote") != NULL)
		    flags |= FLAG_NO_PROMOTE;
		break;
	    case 'd':
		if (optarg == NULL)
		    usage();
		output_directory = optarg;
		break;
	    case 'v':
		version();
		break;	       
	    case 'a': 
	    case 'i':
		{
		    node* newnode =  (node*)malloc(sizeof(node));
		    node** thelist = (c=='a' ? &groups_ok : &groups_ignore);
		    if (!newnode) {
        		fprintf(stderr, "%d@%s: out of memory !\n",
	                    	    __LINE__-1, __FUNCTION__);
			return 1;
		    };
		    if (optarg == NULL)
			usage();
		    
		    if (*thelist) {
			newnode->next = *thelist;
		    } else {
			newnode->next = NULL;
		    };
		    *thelist = newnode;
		    {
			int len = strlen(optarg);
			
			if (optarg[len-1]=='/') {
			    optarg[len-1]='\0';
			    newnode->length = -1;
			} else {
			    newnode->length = len;
			};
			newnode->name = optarg;
		    }
		    break;
		}
	    case 'h':
	    default:
		usage();
		break;
	}
    }

    return 0;
}

/*
 * Returns the user home directory
 */

static char *get_home_dir(void)
{
    char	*user;
    char 	*home = NULL;
    struct passwd *pwd;
    struct stat st;
    int		ret;

    user = getlogin();
    setpwent();
    
    if (user != NULL) {
	pwd = getpwnam(user);
    } else {
	/* oops ... */
	pwd = getpwuid(getuid());
    }

    if (pwd == (struct passwd *)NULL) {
	/* can not identify user home directory ... */
	fprintf(stderr, "Can not identify the user home directory !\n");
	return NULL;
    }
    /* else ... */
    if (pwd->pw_dir != NULL) {
	int maxlen;

	maxlen = strlen(pwd->pw_dir)+strlen(user_confdir)+5;
	if (maxlen > PATH_MAX) {
	    /* whoops, the home dir for our config stuff is too looong */
	    fprintf(stderr, "user home directory %s/%s is too long !\n",
		    home, user_confdir);
	    return NULL;
	}
	home = malloc(maxlen);
	if (home == NULL) {
	    /* oh, no ... memory outrun... */
	    fprintf(stderr, "%d@%s: out of memory !\n",
		    __LINE__-1, __FUNCTION__);
	    endpwent();
	    return NULL;
	}
	snprintf(home, maxlen, "%s/%s",
		 pwd->pw_dir?pwd->pw_dir:"", user_confdir);
    } else
	return NULL;
    endpwent();
    
    ret = stat(home, &st);
    if (ret != 0) {
	/* not having this directory is okay ... */
	return NULL;
    }

    /* okay, it exist... do some more tests */
    if (!S_ISDIR(st.st_mode)) {
	/* but we don't accept a non-directory */
	fprintf(stderr, "%d@%s: %s is not a directory\n",
		__LINE__-1, __FUNCTION__, home);
	return NULL;
    };
    return home;
}

int main (int argc, char **argv)
{
    int ret;
    char *home;
    int have_files = 0;
    
    ProgName = argv[0];
    
    if (parse_options(argc, argv) != 0) {
	fprintf(stderr, "Error parsing program options\n");
	exit(-2);
    }
	
    /* if none or more than one are given, error */
    if (!output_type) {
	fprintf(stderr, "Don't know what configuration to output...\n");
	usage();
	exit (99);
    }
    
    if (sys_confdir == NULL)
	sys_confdir = DEFAULT_SYS_CONFDIR;
    if (user_confdir == NULL)
	user_confdir = DEFAULT_USER_CONFDIR;
    if (root_menu_name == NULL)
	root_menu_name = DEFAULT_ROOT_NAME;
    
    init_packages();
    
    if (!is_set(NO_SYSDIR)) {
	ret = parse_files(sys_confdir);
	if (ret != PARSE_OK)
	    fprintf(stderr, "Parsing system files error: %s\n",
		    parse_errstr(ret));
    }
	
    if (!is_set(NO_USERDIR)) {
	home = get_home_dir();
	if (home != NULL) {
	    ret = parse_files(home);
	    if (ret != PARSE_OK) 
		fprintf(stderr, "Parsing user files error: %s\n",
			parse_errstr(ret));
	    free(home);
	}
    }

#if  HAVE_GNOME
    /* tack on system GNOME files */
    ret = parse_gnome_files("/usr/share/gnome/apps", NULL);
    if (ret != PARSE_OK)
	fprintf(stderr, "Parsing user files error: %s\n", parse_errstr(ret));

    ret = parse_gnome_files("/etc/X11/applnk", NULL);
    if (ret != PARSE_OK)
	fprintf(stderr, "Parsing user files error: %s\n", parse_errstr(ret));
#endif /* HAVE_GNOME */
    
    /* do we have extra files passed on the command line ? */
    while (argc > optind) {
	have_files ++;
	ret = read_file(argv[optind]);
	if (ret != 0)
	    fprintf(stderr, "\tReading file \"%s\" failed !\n",
		    argv[optind]);
	optind++;
    }

    /* Now check to see if we parsed any files */
    if (is_set(NO_SYSDIR) && is_set(NO_USERDIR) && !have_files) {
	fprintf(stderr, "No input files were parsed.\n");
	exit(-1);
    }
    
    /* prepare for the output */
    setbuf(stdout, NULL);
    
    if (build_root(groups_ok, groups_ignore) != 0) {
	fprintf(stderr, "package configuration error.\n");
	exit(-3);
    }

    switch (output_type) {
	case DO_FVWM95:
	    output_fvwm95(root_group);
	    break;
	case DO_FVWM2:
	    output_fvwm2(root_group);
	    break;
	case DO_AFTERSTEP:
	    if (is_set(DIRECTORIES) || is_set(NEWSTYLE_DIRECTORIES))
		output_afterstep_dir(root_group);
	    else
		output_afterstep(root_group);
	    break;
	case DO_MWM:
	    output_mwm(root_group);
	    break;
	case DO_WMAKER:
	    output_wmaker(root_group);
	    break;
	case DO_ICEWM:
	    output_icewm(root_group);
	    break;
	case DO_BLACKBOX:
	    output_blackbox(root_group);
	    break;
	case DO_KDE:
	    output_kde(root_group);
	    break;
	case DO_DEBUG:
	default:    
	    output_debug(root_group);
    }
    return 0;
}

